# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: 2021 Quansight, LLC
# SPDX-FileCopyrightText: 2021 Filipe Laíns <lains@riseup.net>

import os
import subprocess

from typing import Optional

from mesonpy._compat import Collection, Path


class ELF:
    def __init__(self, path: Path) -> None:
        self._path = os.fspath(path)
        self._rpath: Optional[Collection[str]] = None
        self._needed: Optional[Collection[str]] = None

    def _patchelf(self, *args: str) -> str:
        return subprocess.check_output(['patchelf', *args, self._path]).decode()

    @property
    def rpath(self) -> Collection[str]:
        if not self._rpath:
            self._rpath = self._patchelf('--print-rpath').strip().split(';')
        return self._rpath

    @rpath.setter
    def rpath(self, value: Collection[str]) -> None:
        self._patchelf('--set-rpath', ':'.join(value))
        self._rpath = value

    @property
    def needed(self) -> Collection[str]:
        if self._needed is None:
            self._needed = frozenset(self._patchelf('--print-needed').splitlines())
        return self._needed

    @needed.setter
    def needed(self, value: Collection[str]) -> None:
        value = frozenset(value)
        for entry in self.needed:
            if entry not in value:
                self._patchelf('--remove-needed', entry)
        for entry in value:
            if entry not in self.needed:
                self._patchelf('--add-needed', entry)
        self._needed = value



# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: 2021 Quansight, LLC
# SPDX-FileCopyrightText: 2021 Filipe Laíns <lains@riseup.net>

import abc
import re
import sys
import warnings

from typing import Any, Optional

from mesonpy._compat import Literal, Sequence


class Tag(abc.ABC):
    @abc.abstractmethod
    def __init__(self, value: str) -> None: ...

    @abc.abstractmethod
    def __str__(self) -> str: ...

    @property
    @abc.abstractmethod
    def python(self) -> Optional[str]:
        """Python tag."""

    @property
    @abc.abstractmethod
    def abi(self) -> str:
        """ABI tag."""


class StableABITag(Tag):
    _REGEX = re.compile(r'^abi(?P<abi_number>[0-9]+)$')

    def __init__(self, value: str) -> None:
        match = self._REGEX.match(value)
        if not match:
            raise ValueError(f'Invalid PEP 3149 stable ABI tag, expecting pattern `{self._REGEX.pattern}`')
        self._abi_number = int(match.group('abi_number'))

    @property
    def abi_number(self) -> int:
        return self._abi_number

    def __str__(self) -> str:
        return f'abi{self.abi_number}'

    @property
    def python(self) -> Literal[None]:
        return None

    @property
    def abi(self) -> str:
        return f'abi{self.abi_number}'

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and other.abi_number == self.abi_number

    def __hash__(self) -> int:
        return hash(str(self))


class LinuxInterpreterTag(Tag):
    def __init__(self, value: str) -> None:
        parts = value.split('-')
        if len(parts) < 2:
            raise ValueError(
                'Invalid PEP 3149 interpreter tag, expected at '
                f'least 2 parts but got {len(parts)}'
            )

        self._implementation = parts[0]
        self._interpreter_version = parts[1]
        self._additional_information = parts[2:]

        if self.implementation not in ('cpython', 'pypy', 'pypy3'):
            raise NotImplementedError(
                f'Unknown Python implementation: {self.implementation}. '
                'Please report this to https://github.com/FFY00/mesonpy/issues '
                'and include information about the Python distribution you are using.'
            )

    @property
    def implementation(self) -> str:
        return self._implementation

    @property
    def interpreter_version(self) -> str:
        return self._interpreter_version

    @property
    def additional_information(self) -> Sequence[str]:
        return tuple(self._additional_information)

    def __str__(self) -> str:
        return '-'.join((
            self.implementation,
            self.interpreter_version,
            *self.additional_information,
        ))

    @property
    def python(self) -> str:
        if self.implementation == 'cpython':
            # The Python tag for CPython does not seem to include the flags suffixes.
            return f'cp{self.interpreter_version}'.rstrip('dmu')
        elif self.implementation in ('pypy', 'pypy3'):
            interpreter_version = f'{sys.version_info[0]}{sys.version_info[1]}'
            return f'pp{interpreter_version}'
        raise ValueError(f'Unknown implementation: {self.implementation}')

    @property
    def abi(self) -> str:
        # XXX: This is a bit flimsy and needs custom logic to support each case,
        #      but currently there's no better way to do it.
        if self.implementation == 'cpython':
            return f'cp{self.interpreter_version}'
        elif self.implementation == 'pypy':
            return f'pypy_{self.interpreter_version}'
        elif self.implementation == 'pypy3':
            return f'pypy3_{self.interpreter_version}'
        raise ValueError(f'Unknown implementation: {self.implementation}')

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, self.__class__)
            and other.implementation == self.implementation
            and other.interpreter_version == self.interpreter_version
        )

    def __hash__(self) -> int:
        return hash(str(self))


class WindowsInterpreterTag(Tag):
    def __init__(self, value: str) -> None:
        # XXX: This is not actually standardized, so our implementation relies
        #      on observing how the current software behaves!
        self._parts = value.split('-')
        if len(self.parts) != 2:
            warnings.warn(
                'Unexpected native module tag name, the ABI dectection might be broken. '
                'Please report this to https://github.com/FFY00/mesonpy/issues '
                'and include information about the Python distribution you are using.'
            )

    @property
    def parts(self) -> Sequence[str]:
        return tuple(self._parts)

    def __str__(self) -> str:
        return '-'.join(self.parts)

    @property
    def python(self) -> str:
        return self.abi

    @property
    def abi(self) -> str:
        return self._parts[0]

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and other.parts == self.parts

    def __hash__(self) -> int:
        return hash(str(self))



# SPDX-License-Identifier: MIT

import os
import tarfile
import textwrap

import mesonpy

from .conftest import in_git_repo_context


def test_contents(sdist_library):
    sdist = tarfile.open(sdist_library, 'r:gz')

    assert set(sdist.getnames()) == {
        'library-1.0.0/example.c',
        'library-1.0.0/examplelib.c',
        'library-1.0.0/examplelib.h',
        'library-1.0.0/meson.build',
        'library-1.0.0/pyproject.toml',
        'library-1.0.0/PKG-INFO',
    }


def test_contents_subdirs(sdist_subdirs):
    sdist = tarfile.open(sdist_subdirs, 'r:gz')

    assert set(sdist.getnames()) == {
        'subdirs-1.0.0/PKG-INFO',
        'subdirs-1.0.0/meson.build',
        'subdirs-1.0.0/pyproject.toml',
        'subdirs-1.0.0/subdirs/__init__.py',
        'subdirs-1.0.0/subdirs/a/__init__.py',
        'subdirs-1.0.0/subdirs/a/b/c.py',
        'subdirs-1.0.0/subdirs/b/c.py',
    }


def test_contents_unstaged(package_pure, tmpdir):
    new_data = textwrap.dedent('''
    def bar():
        return 'foo'
    ''').strip()

    with open('pure.py', 'r') as f:
        old_data = f.read()

    try:
        with in_git_repo_context():
            with open('pure.py', 'w') as f, open('crap', 'x'):
                f.write(new_data)

            sdist_path = mesonpy.build_sdist(os.fspath(tmpdir))
    finally:
        with open('pure.py', 'w') as f:
            f.write(old_data)
        os.unlink('crap')

    sdist = tarfile.open(tmpdir / sdist_path, 'r:gz')

    assert set(sdist.getnames()) == {
        'pure-1.0.0/PKG-INFO',
        'pure-1.0.0/meson.build',
        'pure-1.0.0/pure.py',
        'pure-1.0.0/pyproject.toml',
    }
    read_data = sdist.extractfile('pure-1.0.0/pure.py').read().replace(b'\r\n', b'\n')
    assert read_data == new_data.encode()



# SPDX-License-Identifier: MIT

import re
import sys

import pytest

import mesonpy._tags


INTERPRETER_VERSION = f'{sys.version_info[0]}{sys.version_info[1]}'


@pytest.mark.parametrize(
    ('value', 'number', 'abi', 'python'),
    [
        ('abi3', 3, 'abi3', None),
        ('abi4', 4, 'abi4', None),
    ]
)
def test_stable_abi_tag(value, number, abi, python):
    tag = mesonpy._tags.StableABITag(value)
    assert str(tag) == value
    assert tag.abi_number == number
    assert tag.abi == abi
    assert tag.python == python
    assert tag == mesonpy._tags.StableABITag(value)


def test_stable_abi_tag_invalid():
    with pytest.raises(ValueError, match=re.escape(
        r'Invalid PEP 3149 stable ABI tag, expecting pattern `^abi(?P<abi_number>[0-9]+)$`'
    )):
        mesonpy._tags.StableABITag('invalid')


@pytest.mark.parametrize(
    ('value', 'implementation', 'version', 'additional', 'abi', 'python'),
    [
        ('cpython-37-x86_64-linux-gnu', 'cpython', '37', ('x86_64', 'linux', 'gnu'), 'cp37', 'cp37'),
        ('cpython-310-x86_64-linux-gnu', 'cpython', '310', ('x86_64', 'linux', 'gnu'), 'cp310', 'cp310'),
        ('cpython-310', 'cpython', '310', (), 'cp310', 'cp310'),
        ('cpython-310-special', 'cpython', '310', ('special',), 'cp310', 'cp310'),
        ('pypy-41', 'pypy', '41', (), 'pypy_41', f'pp{INTERPRETER_VERSION}'),
        ('pypy3-72-x86_64-linux-gnu', 'pypy3', '72', ('x86_64', 'linux', 'gnu'), 'pypy3_72', f'pp{INTERPRETER_VERSION}'),
        ('cpython-310-x86_64-linux-gnu', 'cpython', '310', ('x86_64', 'linux', 'gnu'), 'cp310', 'cp310'),
    ]
)
def test_linux_interpreter_tag(value, implementation, version, additional, abi, python):
    tag = mesonpy._tags.LinuxInterpreterTag(value)
    assert str(tag) == value
    assert tag.implementation == implementation
    assert tag.interpreter_version == version
    assert tag.additional_information == additional
    assert tag.abi == abi
    assert tag.python == python
    assert tag == mesonpy._tags.LinuxInterpreterTag(value)


@pytest.mark.parametrize(
    ('value', 'msg'),
    [
        ('', 'Invalid PEP 3149 interpreter tag, expected at least 2 parts but got 1'),
        ('invalid', 'Invalid PEP 3149 interpreter tag, expected at least 2 parts but got 1'),
    ]
)
def test_linux_interpreter_tag_invalid(value, msg):
    with pytest.raises(ValueError, match=msg):
        mesonpy._tags.LinuxInterpreterTag(value)


@pytest.mark.parametrize(
    ('value', 'parts', 'abi', 'python'),
    [
        ('cp310-win_amd64', ('cp310', 'win_amd64'), 'cp310', 'cp310'),
        ('cp38-win32', ('cp38', 'win32'), 'cp38', 'cp38')
    ]
)
def test_windows_interpreter_tag(value, parts, abi, python):
    tag = mesonpy._tags.WindowsInterpreterTag(value)
    assert str(tag) == value
    assert tag.parts == parts
    assert tag.abi == abi
    assert tag.python == python
    assert tag == mesonpy._tags.WindowsInterpreterTag(value)


@pytest.mark.parametrize('value', ['', 'unknown', 'too-much-information'])
def test_windows_interpreter_tag_warn(value):
    with pytest.warns(Warning, match=(
        'Unexpected native module tag name, the ABI dectection might be broken. '
        'Please report this to https://github.com/FFY00/mesonpy/issues '
        'and include information about the Python distribution you are using.'
    )):
        mesonpy._tags.WindowsInterpreterTag(value)



# SPDX-License-Identifier: MIT

import platform

import pytest

import mesonpy

from .conftest import cd_package


if platform.system() == 'Linux':
    VENDORING_DEPS = {mesonpy._depstr.patchelf}
else:
    VENDORING_DEPS = set()


@pytest.mark.parametrize(
    ('package', 'system_patchelf', 'expected'),
    [
        ('pure', True, set()),  # pure and system patchelf
        ('library', True, set()),  # not pure and system patchelf
        ('pure', False, set()),  # pure and no system patchelf
        ('library', False, VENDORING_DEPS),  # not pure and no system patchelf
    ]
)
def test_get_requires_for_build_wheel(mocker, package, expected, system_patchelf):
    mock = mocker.patch('shutil.which', return_value=system_patchelf)

    if mock.called:  # sanity check for the future if we add another usage
        mock.assert_called_once_with('patchelf')

    with cd_package(package):
        assert set(mesonpy.get_requires_for_build_wheel()) == expected | {
            mesonpy._depstr.wheel,
        }
